/*
 * playlist.cpp
 *
 *  Created on: 18.11.2009
 *      Author: darkstar
 */

#include "playlist.h"
#include "helpers.h"
#include "media.h"
#include "debug.h"

PlayList::PlayList(MediaDatabase *mediaDatabase, const QString dbTable)
	:	QObject(),
	 	mediaDatabase_(mediaDatabase),
	 	dbTable_(dbTable),
		stmtSelectTotalFileSizeAndCountForMediaLocation(NULL),
		stmtSelectAllMediaIDsForMediaLocation(NULL)
{
	connect(mediaDatabase_, SIGNAL(preparingStatements(sqlite3*)),
			this, SLOT(mediaDatabasePreparingStatements(sqlite3*)));

	connect(mediaDatabase_, SIGNAL(finalizingStatements(sqlite3*)),
			this, SLOT(mediaDatabaseFinalizingStatements(sqlite3*)));
}

PlayList::~PlayList()
{

}

void PlayList::beginUpdate()
{
	DENTERMETHOD("PlayList::beginUpdate()");

	if (updateCount_ == 0)
	{
		blockSignals(true);
		mediaDatabase_->beginUpdate();
	}
	++updateCount_;

	DEXITMETHOD("PlayList::beginUpdate()");
}

void PlayList::endUpdate()
{
	DENTERMETHOD("PlayList::endUpdate()");

	--updateCount_;
	if (updateCount_ == 0)
	{
		DTIMERINIT(time);
		mediaDatabase_->endUpdate();
		DTIMERPRINT(time, "mediaDatabase_->endUpdate()");

		DTIMERSTART(time);
		blockSignals(false);
		propagateChanges();
		DTIMERPRINT(time, "change propagation");
	}

	DEXITMETHOD("PlayList::endUpdate()");
}

void PlayList::propagateChanges()
{
	if (updateCount_ == 0)
		emit playListChanged();
}

void PlayList::setDynamicPlayList(bool value)
{
	if (isDynamicPlayList_ != value)
	{
		isDynamicPlayList_ = value;

		if (!isDynamicPlayList_)
			dynPlayListLocations_.clear();

		clear();
	}
}

void PlayList::clear()
{
	mediaDatabase_->clearCache();

	QString query = "DELETE FROM \"" + dbSafeString(dbTable_) + "\";";
	sqlite3_exec(mediaDatabase_->db_, (const char *)query.utf8(), NULL, NULL, NULL);

	propagateChanges();
}

void PlayList::addFile(const QString &fileName)
{

}

void PlayList::addDir(const QString &dirName)
{

}

void PlayList::addURL(const QString &url)
{

}

const QString PlayList::savePlayList(const QString &filename)
{
	QString fullFilename = filename;
	QString extension = QFileInfo(fullFilename).extension(false);

	if (isDynamicPlayList())
	{
		if (extension != "dynplaylist")
			fullFilename += ".dynplaylist";
	}
	else if (extension != "playlist")
		fullFilename += ".playlist";

	QString playListDir = QFileInfo(fullFilename).dirPath(true);

	QFile file(fullFilename);
	if (file.open(IO_WriteOnly))
	{
		QTextStream s(&file);
		s.setEncoding(QTextStream::UnicodeUTF8);

		if (isDynamicPlayList())
		{
			for (QExtStringList::Iterator it = dynPlayList_.begin(); it != dynPlayList_.end(); ++it )
				s << makeLocationRelative(playListDir, *it) << "\n";
		}
		else
		{
			QString query = "SELECT media_id FROM \"" + dbSafeString(dbTable_) + "\" ORDER BY idx, ASC;";

			sqlite3_stmt *vm;
			sqlite3_prepare_v2(mediaDatabase_->db_, query.utf8(), -1, &vm, 0);

			while (sqlite3_step(vm) == SQLITE_ROW)
				s << makeLocationRelative(playListDir, mediaDatabase_->media(sqlite3_column_int(vm, 0))->location().toString()) << "\n";

			sqlite3_finalize(vm);
		}

		// Close File
		file.close();

		return fullFilename;
	}
	else
		return QString::null;
}

void PlayList::mediaDatabasePreparingStatements(sqlite3 *db)
{
	QString query =
		"SELECT SUM(filesize), COUNT() "
		"FROM media, media_location "
		"WHERE media.location_id = media_location.location_id AND media_location.Location = ?1";

	sqlite3_prepare_v2(db, query.utf8(), -1, &stmtSelectTotalFileSizeAndCountForMediaLocation, 0);

	query =
		"SELECT media_id "
		"FROM media, media_location "
		"WHERE media_location.location = ?1 AND media.location_id = media_location.location_id "
		"ORDER BY filename";

	sqlite3_prepare_v2(db, query.utf8(), -1, &stmtSelectAllMediaIDsForMediaLocation, 0);

	query =
		"INSERT INTO \"" + dbSafeString(dbTable_) + "\" "
		"VALUES(?1, ?2, ?3);";

	sqlite3_prepare_v2(db, query.utf8(), -1, &stmtInsertMediaIDIntoPlaylist, 0);
}

void PlayList::mediaDatabaseFinalizingStatements(sqlite3 *db)
{
	sqlite3_finalize(stmtSelectTotalFileSizeAndCountForMediaLocation);
	stmtSelectTotalFileSizeAndCountForMediaLocation = NULL;

	sqlite3_finalize(stmtSelectAllMediaIDsForMediaLocation);
	stmtSelectAllMediaIDsForMediaLocation = NULL;

	sqlite3_finalize(stmtInsertMediaIDIntoPlaylist);
	stmtInsertMediaIDIntoPlaylist = NULL;
}

void PlayList::fillFileList(QExtStringList &filelist, const QString &path, const bool subdir)
{
	QDir dir(path);
	if (!dir.exists()) return;

	dir.setFilter(QDir::Files | QDir::Hidden | (subdir ? QDir::Dirs : 0));
	dir.setSorting(QDir::Unsorted);

	const QFileInfoList *list = dir.entryInfoList();
	QExtString fileName;
	QFileInfoListIterator it(*list);
	QFileInfo *info;

	QExtStringList localFiles;
	uint localFilesTotalSize = 0;

	for (; (info = it.current()); ++it)
	{
		if (info->isFile())
		{
			// only add the file, if its extension is known...
			if (config_.isValidExtension(info->extension(false)))
			{
				localFiles.append(info->absFilePath());
				localFilesTotalSize += info->size();
			}
		}
		else
		{
			fileName = info->fileName();
			if (fileName != "." && fileName != "..")
				fillFileList(filelist, info->absFilePath(), subdir);
		}
	}

	if (localFiles.count() > 0)
	{
#ifdef WINDOWS
		// On Windows canonicalPath() has a really weird and slow behavior (possibly due to QT_GETCWD and QT_CHDIR calls)
		// using absPath() instead...
		QString canonicalPath = dir.absPath();
#else
		QString canonicalPath = dir.canonicalPath();
#endif
		if (canonicalPath.right(1) != '/')
			canonicalPath += '/';

		sqlite3_bind_text(stmtSelectTotalFileSizeAndCountForMediaLocation, 1, canonicalPath.utf8(), -1, SQLITE_TRANSIENT);

		uint filesize = 0;
		uint count = 0;

		if (sqlite3_step(stmtSelectTotalFileSizeAndCountForMediaLocation) == SQLITE_ROW)
		{
			filesize = sqlite3_column_int(stmtSelectTotalFileSizeAndCountForMediaLocation, 0);
			count = sqlite3_column_int(stmtSelectTotalFileSizeAndCountForMediaLocation, 1);
		}

		DHPRINTF("canonicalPath: %s", (const char*)canonicalPath.utf8());
		DHPRINTF("filesize: %d, count: %d", filesize, count);

		sqlite3_reset(stmtSelectTotalFileSizeAndCountForMediaLocation);

		if (filesize == localFilesTotalSize && count == localFiles.count())
		{
			DHPRINTF("using fastpath 1");
			sqlite3_bind_text(stmtSelectAllMediaIDsForMediaLocation, 1, canonicalPath.utf8(), -1, SQLITE_TRANSIENT);

			qHeapSort(localFiles);
			QExtStringList::Iterator it = localFiles.begin();

			while (sqlite3_step(stmtSelectAllMediaIDsForMediaLocation) == SQLITE_ROW)
			{
				QExtString &item = *it;
				item.setID(sqlite3_column_int(stmtSelectAllMediaIDsForMediaLocation, 0));
				DHPRINTF("item: %s, id: %d", (const char *)item.utf8(), item.id());
				++it;
			}

			sqlite3_reset(stmtSelectAllMediaIDsForMediaLocation);
		}
		// TODO: Implement and profile a second fast path attempt, that reads
		// partially available information from the database and assigns media IDs
		// based on this information...

		filelist += localFiles;
	}
}

void PlayList::addFromPlayList(const QExtStringList &filelist, bool interpretPlayList, const QString &playlistDirectory, const QString &dialogTitle)
{
	DTIMERINIT(timer);

	QExtStringList list(filelist);

	// expand various entries in the list if specifically wished for...
	if (interpretPlayList)
	{
		// expand dynamic playlist to real content...
		QExtStringList expandedList;
		for (QExtStringList::ConstIterator it = list.begin(); it != list.end(); ++it)
		{
			QString location(*it);

			if (location.startsWith("#") || location.startsWith(";"))
				continue;

			// WHY OH WHY is there startsWith but no endsWith in QString in Qt 2.3.x ?!?!
			if (location.right(3) == "/**" || location == "**")
			{
				// recursive scanning of location...
				location = resolveLocation(playlistDirectory, location.left(QMAX(0, location.length() - 3)));
				QExtStringList fileList;
				fillFileList(fileList, location, true);
				qHeapSort(fileList);
				expandedList += fileList;
			}
			else if (location.right(2) == "/*" || location.right(1) == "/")
			{
				// all files of the location...
				location = resolveLocation(playlistDirectory, location.left(location.findRev('/')));
				QExtStringList fileList;
				fillFileList(fileList, location, false);
				qHeapSort(fileList);
				expandedList += fileList;
			}
			else if (location.startsWith("http://") || location.startsWith("mms://"))
				expandedList.append(location);
			else
			{
				location = resolveLocation(playlistDirectory, location);
				QFileInfo fi(location);

				if (fi.exists())
				{
					if (fi.isDir())
					{
						QExtStringList fileList;
						fillFileList(fileList, location, false);
						qHeapSort(fileList);
						expandedList += fileList;
					}
					else if (config_.isValidExtension(fi.extension()))
						// just add the (file) location to the expanded list...
						expandedList.append(location);
				}
			}
		}

		list = expandedList;
	}

	// start reading the playlist entries...
	beginUpdate();

	QTime time;
	time.start();

	sqlite3_stmt *vm = stmtInsertMediaIDIntoPlaylist;

	for (QExtStringList::ConstIterator it = list.begin(); it != list.end(); ++it)
	{
		const QExtString &location = *it;

		unsigned long mediaID = NULL;

		if (location.id())
		{
			DHPRINTF("ID %d set on %s", location.id(), (const char *)location.utf8());
			mediaID = location.id();
		}
		else
		{
			// skip comments
			if (location.startsWith("#") || location.startsWith(";"))
				continue;

			DHPRINTF("No ID set on %s", (const char *)location.utf8());

			QString resolvedLoc;

			if (!(isPathAbsolute(location) || location.startsWith("http://") || location.startsWith("mms://")))
				resolvedLoc = resolveLocation(playlistDirectory, location);
			else
				resolvedLoc = location;

			mediaID = mediaDatabase_->getMediaIDForLocation(MediaLocation(resolvedLoc));
		}

		sqlite3_bind_null(vm, 1);
		sqlite3_bind_null(vm, 2);
		sqlite3_bind_int(vm, 3, mediaID);

		sqlite3_step(vm);

		sqlite3_reset(vm);
	}

	DTIMERPRINT(timer, "playlist added to DB in");

	endUpdate();
	propagateChanges();

	DMEMSTAT();
}
